import sys
import time

try:
    import queue
except ImportError:
    import Queue as queue
from acq4.util.Thread import Thread


class PatchThread(Thread):
    """Background thread that acquires and runs multiple jobs for a single pipette
    """

    class Stopped(Exception):
        """Raised when we request the thread to stop
        """

    def __init__(self, dev, module):
        Thread.__init__(self, name=f"VoltageImaging_PatchThread_{dev.name()}")
        self.dev = dev
        self.jobQueue = module.jobQueue
        self.module = module
        self._stop = False

    def stop(self):
        self._stop = True

    def start(self):
        self._stop = False
        Thread.start(self)

    def checkStop(self):
        if self._stop is True:
            raise self.Stopped()

    def run(self):
        while self._stop is False:
            if self.dev.waitingForSwap or self.dev.active is False:
                # wait for user to swap pipette or enable
                time.sleep(0.5)
                continue

            pa = self.jobQueue.requestJob(self.dev)
            self._currentPatchAttempt = pa
            if pa is None:
                # no jobs right now; sleep and try again.

                time.sleep(3)
                continue

            try:
                pa.startLogging()
                pa.setStatus(f"start patch protocol: {pa.protocol.name}")
                protocol = pa.protocol(self, pa)
                protocol.runPatchProtocol()
                if pa.error is None:
                    pa.setStatus("success")
                else:
                    pa.setStatus(f"failed: {str(pa.error[1])}")
            except self.Stopped:
                protocol.abortPatchProtocol()
                pa.setStatus("aborted")
                return
            except Exception as exc:
                pa.setError(sys.exc_info())
            finally:
                pa.stopLogging()



from collections import OrderedDict

from . import recalibrate, test, task_runner, mock
from .patch_protocol import PatchProtocol


def allPatchProtocols(rootClass=PatchProtocol):
    prots = OrderedDict()
    for cls in rootClass.__subclasses__():
        if cls.name is not None:
            prots[cls.name] = cls
        prots.update(allPatchProtocols(cls))
    return prots



import threading
import time

import numpy as np
import pyqtgraph as pg
from acq4.util import Qt
from acq4.util.threadrun import runInGuiThread

from .patch_protocol import PatchProtocol


class RecalibrateProtocol(PatchProtocol):
    """Base class for protocols that visit each target +10um and measure the pipette calibration error 
    """

    name = None

    def __init__(self, patchThread, patchAttempt):
        PatchProtocol.__init__(self, patchThread, patchAttempt)
        self.dev = patchThread.dev
        self.module = patchThread.module
        self.clickEvent = threading.Event()
        self.stageCameraLock = self.module.stageCameraLock
        self.camera = self.module.getCameraDevice()
        self.cameraMod = self.module.getCameraModule()

    def runPatchProtocol(self):
        # How far above target to run calibration?
        #  - low values (10 um) potentially have worse machine vision performance due to being very close to cells
        #  - high values (100 um) potentially yield a bad correction due to errors that accumulate over the large distance to the cell
        calibrationHeight = 30e-6

        pa = self.patchAttempt
        if not hasattr(pa, "originalPosition"):
            pa.originalPosition = np.array(pa.position)

        # move to 100 um above current position
        pos = self.dev.pipetteDevice.globalPosition()
        pos[2] += 100e-6
        fut = self.dev.pipetteDevice._moveToGlobal(pos, "fast")
        self.wait([fut])

        # move to 100 um above target z value
        pos = pa.pipetteTargetPosition()
        pos[2] += 100e-6
        fut = self.dev.pipetteDevice._moveToGlobal(pos, "fast")
        self.wait([fut])

        # set pipette target position
        self.dev.pipetteDevice.setTarget(pa.pipetteTargetPosition())

        # move pipette to 10 um above corrected target
        pipPos = pa.pipetteTargetPosition() + np.array([0, 0, calibrationHeight])
        # don't use target move here; we don't need all the obstacle avoidance.
        # fut = self.dev.pipetteDevice.goTarget(speed='fast')
        pfut = self.dev.pipetteDevice._moveToGlobal(pipPos, speed="slow")

        with self.stageCameraLock.acquire() as fut:
            pa.setStatus("Waiting for stage/camera")
            self.wait([fut], timeout=None)

            # move stage/focus above actual target
            camPos = pa.globalTargetPosition() + np.array([0, 0, calibrationHeight])
            cfut = self.camera.moveCenterToGlobal(camPos, "fast")
            self.wait([pfut, cfut], timeout=None)

            # Offset from target to where pipette actually landed
            try:
                self.patchAttempt.pipetteError = self.getPipetteError()
            except RuntimeError:
                self.patchAttempt.pipetteError = np.array([np.nan] * 3)
                raise

    def getPipetteError(self):
        """Return the calibration offset between the expected pipette position and the known pipette position
        """
        raise NotImplementedError()


class AutoRecalibrateProtocol(RecalibrateProtocol):
    name = "auto recalibrate"

    def __init__(self, *args, **kwds):
        RecalibrateProtocol.__init__(self, *args, **kwds)
        self.line = None

    def getPipetteError(self):
        """Return error vector that should be added to pipette position fotr the current target.

        Error vector may contain NaN to indicate that the correction failed and this target should not be attempted.
        """
        pa = self.patchAttempt
        pa.setStatus("Measuring pipette error")

        perfVals = []
        pipetteDiffVals = []
        targetErrVals = []
        focusErrVals = []

        targetPos = np.array(pa.pipetteTargetPosition())

        # Make a few attempts to optimize pipette position. Iterate until
        #  - z is in focus on the pipette tip
        #  - pipette x,y is over the target
        for i in range(4):
            cameraPos = self.camera.globalCenterPosition("roi")

            # pipette position according to manipulator
            reportedPos = np.array(self.dev.pipetteDevice.globalPosition())

            # estimate tip position measured by machine vision
            measuredPos, perf = self.dev.pipetteDevice.tracker.measureTipPosition(threshold=0.4, movePipette=False)
            measuredPos = np.array(measuredPos)

            # generate some error metrics:
            # how far is the pipette from its reported position
            pipetteDiff = measuredPos - reportedPos
            # how far in Z is the pipette from the focal plane
            focusError = abs(measuredPos[2] - cameraPos[2])
            # how far in XY is the pipette from the target
            targetDiff = targetPos[:2] - measuredPos[:2]
            targetError = np.linalg.norm(targetDiff)

            # track performance so we can decide later whether to abandon this point
            perfVals.append(perf)
            pipetteDiffVals.append(pipetteDiff)
            focusErrVals.append(focusError)
            targetErrVals.append(targetError)

            # show the error line and pause briefly (just for debugging; we could remove this to speed up the process)
            self.showErrorLine(reportedPos, measuredPos)

            futs = []
            if focusError > 3e-6:
                # refocus on pipette tip (don't move pipette in z because if error prediction is wrong, we could crash)
                cameraPos[2] = measuredPos[2]
                futs.append(self.camera.moveCenterToGlobal(cameraPos, "slow"))

            if targetError > 1.5e-6:
                # reposition pipette x,y closer to target
                ppos = reportedPos.copy()
                ppos[:2] += targetDiff
                futs.append(self.dev.pipetteDevice._moveToGlobal(ppos, "slow"))

            if len(futs) > 0:
                # wait for requested moves to complete and try again
                self.wait(futs)
                time.sleep(0.3)  # wait for positions to catch up.. we can remove this after bug fixed!
                pa.setStatus(f"Measuring pipette error: adjust and iterate  ({i:d})")
            else:
                # no moves needed this round; we are done.
                break

        # Now decide whether to pass or fail this calibration.
        if focusErrVals[-1] > 3e-6 or targetErrVals[-1] > 3e-6 or perfVals[-1] < 0.5:
            raise RuntimeError(
                f"Measuring pipette error: failed  (focus error: {focusErrVals}  target error: {targetErrVals}  correlation: {perfVals})"
            )

        pa.setStatus(f"Measuring pipette error: success {pipetteDiff}")
        return pipetteDiff

    def showErrorLine(self, pt1, pt2):
        runInGuiThread(self._showErrorLine, pt1, pt2)
        time.sleep(1.5)
        runInGuiThread(self._removeErrorLine)

    def _showErrorLine(self, pt1, pt2):
        self._removeErrorLine()
        self.line = pg.QtGui.QGraphicsLineItem(pt1[0], pt1[1], pt2[0], pt2[1])
        self.line.setPen(pg.mkPen("r"))
        self.cameraMod.window().addItem(self.line)

    def _removeErrorLine(self):
        if self.line is None:
            return
        self.line.scene().removeItem(self.line)
        self.line = None


class ManualRecalibrateProtocol(RecalibrateProtocol):
    name = "manual recalibrate"

    def __init__(self, patchThread, patchAttempt):
        RecalibrateProtocol.__init__(self, patchThread, patchAttempt)

    def runPatchProtocol(self):
        # Grab click events fom the camera module while this protocol is running
        self.cameraMod.window().getView().scene().sigMouseClicked.connect(
            self.cameraModuleClicked, Qt.Qt.DirectConnection
        )
        try:
            RecalibrateProtocol.runPatchProtocol(self)
        finally:
            self.cameraMod.window().getView().scene().sigMouseClicked.disconnect(self.cameraModuleClicked)

    def getPipetteError(self):
        pa = self.patchAttempt
        pa.setStatus("Waiting for user click")
        clickPos = self.getClickPosition()
        pos = np.array(self.dev.pipetteDevice.globalPosition())
        return clickPos - pos

    def getClickPosition(self):
        self.clickEvent.clear()
        while True:
            if self.clickEvent.wait(0.2):
                break
            self.checkStop()
        return np.array(self.lastClick)

    def cameraModuleClicked(self, ev):
        cameraPos = self.camera.mapToGlobal([0, 0, 0])
        globalPos = self.cameraMod.window().getView().mapSceneToView(ev.scenePos())
        globalPos = [globalPos.x(), globalPos.y(), cameraPos[2]]
        self.lastClick = globalPos
        self.clickEvent.set()



"""
ACQ4's built in pipette motion planning takes extra steps to avoid collisions with upright microscope objectives. 
On the inverted scope, we don't need to avoid the objective.
"""
from acq4.devices.Pipette import Pipette
from acq4.devices.Pipette.planners import PipetteMotionPlanner


class TargetMotionPlanner(PipetteMotionPlanner):
    """Move directly to target
    """

    def _move(self):
        pip = self.pip
        speed = self.speed
        target = pip.targetPosition()
        return pip._moveToGlobal(target, speed=speed)


Pipette.defaultMotionPlanners["target"] = TargetMotionPlanner


class ApproachMotionPlanner(PipetteMotionPlanner):
    """Move directly to approach position
    """

    def _move(self):
        pip = self.pip
        speed = self.speed
        target = pip.targetPosition()
        target[2] = self.pip.approachDepth()
        return pip._moveToGlobal(target, speed=speed)


Pipette.defaultMotionPlanners["approach"] = ApproachMotionPlanner



# from acq4.devices.PatchPipette.states import PatchPipetteCleanState
# from acq4.devices.PatchPipette.statemanager import PatchPipetteStateManager

# implementation moved to mainline clean state
# (but leaving this here as an example of implementing custom states)

# class PatchPipetteCleanState(PatchPipetteCleanState):
#     """Customization of patch pipette cleaning state with motion control for recording chamber.
#     """
#     def run(self):
#         self.lastApproachPos = None
#         PatchPipetteCleanState.run(self)

#     def gotoApproachPosition(self, pos):
#         """
#         """
#         dev = self.dev
#         currentPos = dev.pipetteDevice.globalPosition()

#         # first move back in x and up in z, leaving y unchanged
#         print('approachHeight:', self.config['approachHeight'])
#         approachPos1 = [pos[0], currentPos[1], pos[2] + self.config['approachHeight']]
#         fut = dev.pipetteDevice._moveToGlobal(approachPos1, 'fast')
#         self.waitFor(fut)
#         if self.resetPos is None:
#             self.resetPos = approachPos1

#         # now move y over the well
#         approachPos2 = [pos[0], pos[1], pos[2] + self.config['approachHeight']]
#         fut = dev.pipetteDevice._moveToGlobal(approachPos2, 'fast')
#         self.lastApproachPos = approachPos2
#         self.waitFor(fut)


# # Install customized state class into default state list
# PatchPipetteStateManager.stateHandlers['clean'] = PatchPipetteCleanState



import sys
import time

import numpy as np

try:
    import queue
except ImportError:
    import Queue as queue
from acq4.Manager import getManager
from acq4.util import Qt
from acq4.util.threadrun import runInGuiThread
from .patch_protocol import PatchProtocol


class TaskRunnerPatchProtocol(PatchProtocol):
    """Patch protocol implementing:

    - Move to cell, take brightfield photo, autopatch 
    - Initiate TaskRunner protocol
    - Clean pipette
    - Move pipette home and request swap (if broken / clogged)
    """

    name = "task runner"

    def __init__(self, patchThread, patchAttempt):
        PatchProtocol.__init__(self, patchThread, patchAttempt)
        self.dev = patchThread.dev
        self.module = patchThread.module
        self.stageCameraLock = self.module.stageCameraLock
        self.camera = self.module.getCameraDevice()
        self.scope = self.camera.getScopeDevice()

        man = getManager()
        self.dh = man.getCurrentDir().mkdir(f"patch_attempt_{self.patchAttempt.pid:04d}", autoIncrement=True)
        patchAttempt.setLogFile(self.dh["patch.log"])

        self.stateQueue = queue.Queue()
        # this code is running in a thread, so it is necessary to specify that
        # the signal must be delivered in the main thread (since we are not running an event loop)
        self.dev.stateManager().sigStateChanged.connect(self.devStateChanged, Qt.Qt.DirectConnection)

    def devStateChanged(self, stateManager, state):
        self.stateQueue.put(state)

    def runPatchProtocol(self):
        pa = self.patchAttempt

        if not self.dev.isTipClean():
            self.cleanPipette()

        try:
            self.dev.setState("bath")
            time.sleep(5)

            self.patchCell()

            finalState = self.dev.getState()
            if finalState.stateName != "whole cell":
                raise Exception(f"Failed to reach whole cell state (ended at {finalState}).")

            with self.stageCameraLock.acquire() as fut:
                pa.setStatus("Waiting for stage/camera")
                self.wait([fut], timeout=None)
                self.configureCamera()
                self.runProtocol(pa)

        except:
            pa.setError(sys.exc_info())
        finally:
            if self.dev.broken:
                self.swapPipette()
            elif not self.dev.clean:
                self.cleanPipette()

    def patchCell(self):
        pa = self.patchAttempt

        # Set target cell position, taking error correction into account
        targetPos = pa.pipetteTargetPosition()
        if not np.all(np.isfinite(targetPos)):
            raise Exception("No valid target position for this attempt (probably automatic recalibration failed)")

        pa.setStatus("moving to target")
        self.dev.pipetteDevice.setTarget(targetPos)

        # move to 100 um above cell, fast
        pos = np.array(targetPos) + np.array([100e-6, 100e-6, 100e-6])
        fut = self.dev.pipetteDevice._moveToGlobal(pos, speed="fast")
        self.wait([fut])

        # move to 10 um above cell, slow
        pos = np.array(targetPos) + np.array([0, 0, 10e-6])
        # don't use target move here; we don't need all the obstacle avoidance.
        # fut = self.dev.pipetteDevice.goTarget(speed='fast')
        fut = self.dev.pipetteDevice._moveToGlobal(pos, speed="slow")
        self.wait([fut])

        self.clearStateQueue()

        # kick off cell detection; wait until patched or failed
        pa.setStatus("cell patching")
        self.dev.setState("cell detect")
        while True:
            self.checkStop()
            try:
                state = self.stateQueue.get(timeout=0.2)
            except queue.Empty:
                continue

            if state.stateName in ("whole cell", "fouled", "broken"):
                return
            else:
                pa.setStatus(f"cell patching: {state.stateName}")

            while True:
                try:
                    # raise exception if this state fails
                    state.wait(timeout=0.2)
                    break
                except state.Timeout:
                    self.checkStop()

    def abortPatchProtocol(self):
        pass

    def clearStateQueue(self):
        # clear out information about any pipette states before now
        while not self.stateQueue.empty():
            self.stateQueue.get(timeout=0)

    def runProtocol(self, pa):
        """Cell is patched; lock the stage and begin protocol.
        """
        # focus camera on cell
        pa.setStatus("focus on cell")
        self.camera.moveCenterToGlobal(pa.globalTargetPosition(), speed="fast", center="roi").wait()

        man = getManager()
        turret = man.getDevice("FilterTurret")
        illum = man.getDevice("Illumination")

        # set filter wheel / illumination
        turret.setPosition(1).wait()
        time.sleep(2)  # scope automatically changes RL/TL settings, sometimes in a bad way. sleep and set manually:
        illum.SetTLIllumination(1)
        illum.SetRLIllumination(1)

        # take a picture
        pa.setStatus("say cheese!")
        frame = self.camera.acquireFrames(n=1, stack=False)
        frame.saveImage(self.dh, "patch_image.tif")

        pa.setStatus("running whole cell protocol")

        # switch to RL
        turret.setPosition(0).wait()
        time.sleep(2)  # scope automatically changes RL/TL settings, sometimes in a bad way. sleep and set manually:
        illum.SetTLIllumination(2)
        illum.SetRLIllumination(2)
        time.sleep(1)

        try:
            # take another picture
            cameraParams = self.camera.getParams()
            self.camera.setParams({"exposure": 0.05, "binning": (4, 4)})

            frame = self.camera.acquireFrames(n=1, stack=False)
            frame.saveImage(self.dh, "fluor_image.tif")

            man = getManager()
            # TODO: select correct task runner for this pipette
            taskrunner = None
            for mod in man.listModules():
                if not mod.startswith("Task Runner"):
                    continue
                mod = man.getModule(mod)
                if self.dev.clampDevice.name() in mod.docks:
                    taskrunner = mod
                    break

            assert taskrunner is not None, f"No task runner found that uses {self.dev.clampDevice.name()}"

            # 300 Hz
            # self.camera.setParams({'regionH': 700, 'regionY': 680, 'regionX': 8, 'regionW': 2028, 'exposure': 0.0030013})
            # 1kHz
            self.camera.setParams(
                {
                    "regionH": 164,
                    "regionY": 940,
                    "regionX": 8,
                    "regionW": 2032,
                    "exposure": 0.0010134,
                    "binning": (4, 4),
                }
            )

            # prepare camera to be triggered by the DAQ for this pipette
            self.configureCamera()
            fut = runInGuiThread(taskrunner.runSequence, store=True, storeDirHandle=self.dh)
            try:
                self.wait([fut], timeout=300)
            except self.patchThread.Stopped:
                fut.stop()
                raise

        finally:
            # switch off RL
            turret.setPosition(1).wait()
            time.sleep(2)  # scope automatically changes RL/TL settings, sometimes in a bad way. sleep and set manually:
            illum.SetTLIllumination(1)
            illum.SetRLIllumination(1)

            self.camera.setParams(cameraParams)  # , autoRestart=True, autoCorrect=True)

            pa.setStatus("restart acquire video of camera")
            self.camera.start()

        time.sleep(2)
        pa.setStatus("whole cell protocol complete")

    def configureCamera(self):
        """Set camera exposure/trigger channels for this pipette's DAQ.
        """
        # note: we'd love it if the camera and DAQ could just automatically decide which trigger
        # channels to use, but that's not supported yet so this is a temporary workaround.
        if "cameraChannels" in self.module.config:
            exp, trig = self.module.config["cameraChannels"][self.dev.name()]
            self.camera.reconfigureChannel("exposure", {"channel": exp})
            self.camera.reconfigureChannel("trigger", {"channel": trig})

    def cleanPipette(self):
        pa = self.patchAttempt
        pa.setStatus("cleaning pipette")
        self.clearStateQueue()
        fut = self.dev.setState("clean")

        # wait for cleaning to finish
        self.wait([fut], timeout=120)

    def swapPipette(self):
        pa = self.patchAttempt
        pa.setStatus("requesting new pipette")
        self.dev.setState("out")
        self.dev.goHome("fast")
        self.dev.requestNewPipette()



import time


class PatchProtocol(object):
    """Base class for defining actions to perform once a cell+pipette have been selected and are
    ready for patching.

    This class may be customized to provide support for new experiment types.

    Might include actions like:
    - Move pipette to cell
    - Take brightfield image
    - Attempt patch
    - Initiate reocrding protocol
    - Various abort procedures
    - Clean / swap pipette
    """

    name = None

    def __init__(self, patchThread, patchAttempt):
        self.patchThread = patchThread
        self.patchAttempt = patchAttempt

    def runPatchProtocol(self):
        """Implements (in subclass) the sequence of actions defining this protocol.

        This method should call self.checkStop() periodically in order to respond
        to abort requests.
        """
        raise NotImplementedError()

    def abortPatchProtocol(self):
        """This method is called when the protocol has stopped early; may be reimplemented
        to perform cleanup.
        """
        raise NotImplementedError()

    def checkStop(self):
        """Raise an exception if an abort was requested.
        """
        self.patchThread.checkStop()

    def lock(self, lock, timeout=20.0):
        """Return a context manager that attempts to lock a mutex while checking for abort requests.
        """
        return Locker(self, lock, timeout)

    def wait(self, futures, timeout=20.0):
        """Wait for multiple futures to complete while also checking for abort requests.
        """
        if len(futures) == 0:
            return
        start = time.time()
        while True:
            self.checkStop()
            allDone = True
            for fut in futures[:]:
                try:
                    fut.wait(0.2)
                    futures.remove(fut)
                except fut.Timeout:
                    allDone = False
                    break
            if allDone:
                break
            if timeout is not None and time.time() - start > timeout:
                raise futures[0].Timeout(f"Timed out waiting for {futures!r}")


class Locker(object):
    """Mutex locker that periodically checks for protocol abort requests.
    """

    class Timeout(Exception):
        pass

    def __init__(self, protocol, lock, timeout=20.0):
        self.protocol = protocol
        self.lock = lock
        self.timeout = timeout
        self.unlock = False

    def __enter__(self):
        start = time.time()
        while True:
            if self.lock.tryLock(200):
                self.unlock = True
                self.protocol.checkStop()
                return self
            if self.timeout is not None and time.time() - start > self.timeout:
                raise Locker.Timeout("Timed out waiting for lock")

    def __exit__(self, *args):
        if self.unlock:
            self.lock.unlock()



import json
import traceback
from collections import OrderedDict

import numpy as np
import pyqtgraph as pg
from acq4.util import Qt
from pyqtgraph import ptime


class PatchAttempt(Qt.QObject):
    """Stores 3D location, status, and results for a point to be patched.
    """

    statusChanged = Qt.Signal(object, object)  # self, status
    newEvent = Qt.Signal(object, object)  # self, event

    def __init__(self, pid, position, treeItem, targetItem):
        Qt.QObject.__init__(self)
        self.assigned_protocols = set()
        self.pid = pid
        self.position = position
        self.pipetteError = None
        self.treeItem = treeItem
        self.targetItem = targetItem
        self.protocol = None
        self.pipette = None
        self.status = None
        self.result = {}
        self.error = None
        self.log = []
        self.logFile = None

    def reset(self):
        self.stopLogging()
        self.assigned_protocols = set()
        self.pipette = None
        self.pipetteError = None
        self.setStatus("reset")
        self.result = {}
        self.error = None
        self.log = []
        self.status = None

    def hasStarted(self):
        return self.status is not None

    def setProtocol(self, prot):
        self.assigned_protocols.add(prot.name)
        self.protocol = prot

    def setStatus(self, status):
        self.status = status
        self.log.append(
            OrderedDict(
                [
                    ("device", "None" if self.pipette is None else self.pipette.name()),
                    ("event_time", ptime.time()),
                    ("event", "statusChanged"),
                    ("status", status),
                ]
            )
        )
        self.statusChanged.emit(self, status)

    def assignPipette(self, pip):
        assert self.pipette in (None, pip), "Pipette can only be assigned once"
        self.pipette = pip
        self.setStatus("assigned")

    def pipetteEvent(self, pip, event):
        self.log.append(event)
        self.writeLogEvent(event)
        self.newEvent.emit(self, event)

    def setLogFile(self, fh):
        self.logFile = fh
        for ev in self.log:
            self.writeLogEvent(ev)

    def writeLogEvent(self, event):
        if self.logFile is None:
            return
        try:
            ev = json.dumps(event)
        except Exception:
            print(repr(ev))
            raise
        with open(self.logFile.name(), "a") as fh:
            fh.write(ev)
            fh.write("\n")

    def setError(self, excinfo):
        self.error = excinfo
        exclass, exc, tb = excinfo
        self.setStatus(f'error during "{self.status}" : {str(exc)}')
        ev = OrderedDict(
            [
                ("device", "None" if self.pipette is None else self.pipette.name()),
                ("event_time", ptime.time()),
                ("event", "error"),
                ("error", traceback.format_exception(*excinfo)),
            ]
        )
        self.pipetteEvent(self.pipette, ev)

    def pipetteTargetPosition(self):
        """Return the global coordinate of the selected target for this patch attempt, corrected
        for the pipette position error.
        """
        pos = np.array(self.position)
        if self.pipetteError is not None:
            pos -= self.pipetteError
        return pos

    def globalTargetPosition(self):
        """Return the global coordinate of the selected target for this patch attempt.
        """
        return np.array(self.position)

    def startLogging(self):
        """Connect device signals to begin logging events.
        """
        self.stopLogging()
        self.pipette.sigNewEvent.connect(self.pipetteEvent)

    def stopLogging(self):
        """Disconnect device signals to stop logging events.
        """
        if self.pipette is None:
            return
        pg.disconnect(self.pipette.sigNewEvent, self.pipetteEvent)

    def formatLog(self):
        """Return a string describing all events logged for this attempt (for debugging)
        """
        log = [
            "========================================",
            f"       Patch attempt {self.pid:d}",
            f"       Current status: {self.status}",
            "========================================",
            "Event log:",
        ]
        for event in self.log:
            log.append("  ".join([f"{k}={v}" for k, v in event.items()]))
        if self.error is not None:
            log.append("========================================")
            log.append("Error:")
            log.extend(traceback.format_exception(*self.error))
        return "\n".join(log)



import threading

import numpy as np
import scipy.stats
from acq4.util import Qt

from .patch_protocol import PatchProtocol


class ResultText(Qt.QPlainTextEdit):
    updateText = Qt.Signal()

    def __init__(self):
        Qt.QPlainTextEdit.__init__(self)
        self.records = []
        self.resetBtn = Qt.QPushButton("reset")
        self.resetBtn.setParent(self)
        self.resetBtn.resize(40, 15)
        self.resetBtn.clicked.connect(self.reset)
        self.updateText.connect(self._updateText)
        self.setWindowTitle("Stage/pipette accuracy test results")
        doc = self.document()
        font = doc.defaultFont()
        font.setFamily("Monospace")
        font.setStyleHint(font.TypeWriter)
        doc.setDefaultFont(font)
        self.resize(600, 600)

    def addRecord(self, rec):
        self.records.append(rec)
        self.updateText.emit()

    def _updateText(self):
        report = "\n\nPipette  X         Y         Z         XY        Stage\n"

        # per-target errors
        for rec in self.records:
            fields = {
                "stage": f"{np.linalg.norm(rec['stage']) * 1e6:0.2f}um",
                "pipettex": f"{rec['pipette'][0] * 1e6:0.2f}um",
                "pipettey": f"{rec['pipette'][1] * 1e6:0.2f}um",
                "pipettez": f"{rec['pipette'][2] * 1e6:0.2f}um",
                "pipettexy": f"{np.linalg.norm(rec['pipette'][:2]) * 1e6:0.2f}um",
            }
            report += "         {pipettex:10s}{pipettey:10s}{pipettez:10s}{pipettexy:10s}{stage:10s}\n".format(**fields)

        pip_err = np.array([rec["pipette"] for rec in self.records])
        stage_err = np.array([rec["stage"] for rec in self.records])

        # summary stats
        fields = {
            "stdx": f"{pip_err[:, 0].std() * 1e6:0.2f}um",
            "stdy": f"{pip_err[:, 1].std() * 1e6:0.2f}um",
            "stdz": f"{pip_err[:, 2].std() * 1e6:0.2f}um",
            "stdxy": f"{np.linalg.norm([pip_err[:, 0].std(), pip_err[:, 1].std()]) * 1e6:0.2f}um",
        }
        report += "Std      {stdx:10s}{stdy:10s}{stdz:10s}{stdxy:10s}\n".format(**fields)
        fields = {
            "meanx": f"{pip_err[:, 0].mean() * 1e6:0.2f}um",
            "meany": f"{pip_err[:, 1].mean() * 1e6:0.2f}um",
            "meanz": f"{pip_err[:, 2].mean() * 1e6:0.2f}um",
            "meanxy": f"{np.linalg.norm([pip_err[:, 0].mean(), pip_err[:, 1].mean()]) * 1e6:0.2f}um",
        }
        report += "Mean     {meanx:10s}{meany:10s}{meanz:10s}{meanxy:10s}\n".format(**fields)

        xydist = np.linalg.norm(pip_err[:, :2], axis=1)
        report += f"XY distance 95th percentile: {scipy.stats.scoreatpercentile(xydist, 95.0):0.2f}um"

        self.document().setPlainText(report)
        self.show()

    def reset(self):
        self.records = []
        self.document().setPlainText("")


resultText = ResultText()


class TestPatchProtocol(PatchProtocol):
    """Simplified patch protocol used for testing pipette / stage movement
    """

    name = "stage/manipulator test"

    def __init__(self, patchThread, patchAttempt):
        PatchProtocol.__init__(self, patchThread, patchAttempt)
        self.dev = patchThread.dev
        self.module = patchThread.module
        self.clickEvent = threading.Event()
        self.stageCameraLock = self.module.stageCameraLock
        self.camera = self.module.getCameraDevice()
        self.cameraMod = self.module.getCameraModule()
        self.lines = None

    def runPatchProtocol(self):
        # Grab click events fom the camera module while this protocol is running
        self.cameraMod.window().getView().scene().sigMouseClicked.connect(
            self.cameraModuleClicked, Qt.Qt.DirectConnection
        )
        try:
            self._runPatchProtocol()
        finally:
            self.cameraMod.window().getView().scene().sigMouseClicked.disconnect(self.cameraModuleClicked)

    def getClickPosition(self):
        self.clickEvent.clear()
        while True:
            if self.clickEvent.wait(0.2):
                break
            self.checkStop()
        return np.array(self.lastClick)

    def cameraModuleClicked(self, ev):
        cameraPos = self.camera.mapToGlobal([0, 0, 0])
        globalPos = self.cameraMod.window().getView().mapSceneToView(ev.scenePos())
        globalPos = [globalPos.x(), globalPos.y(), cameraPos[2]]
        self.lastClick = globalPos
        self.clickEvent.set()

    def _runPatchProtocol(self):
        pa = self.patchAttempt
        pa.setStatus("moving to target")

        # move to 100 um above current position
        pos = self.dev.pipetteDevice.globalPosition()
        pos[2] += 100e-6
        fut = self.dev.pipetteDevice._moveToGlobal(pos, "fast")
        self.wait([fut])

        # move to 100 um above target z value
        pos = pa.pipetteTargetPosition()
        pos[2] += 100e-6
        fut = self.dev.pipetteDevice._moveToGlobal(pos, "fast")
        self.wait([fut])

        self.dev.pipetteDevice.setTarget(pa.pipetteTargetPosition())

        # move to 10 um above cell
        pipPos = np.array(pa.pipetteTargetPosition()) + np.array([0, 0, 10e-6])
        # don't use target move here; we don't need all the obstacle avoidance.
        # fut = self.dev.pipetteDevice.goTarget(speed='fast')
        pfut = self.dev.pipetteDevice._moveToGlobal(pipPos, speed="slow")

        with self.stageCameraLock.acquire() as fut:
            pa.setStatus("Waiting for stage/camera")
            self.wait([fut], timeout=None)
            # Move to actual target, wait for click
            camPos = pa.globalTargetPosition()
            cfut = self.camera.moveCenterToGlobal(camPos, "fast")
            self.wait([pfut, cfut], timeout=None)

            pa.setStatus("Waiting click on target")
            targetClickPos = self.getClickPosition()
            stageErr = targetClickPos - camPos

            # Move to target + 10um, wait for click on pipette
            cfut = self.camera.moveCenterToGlobal(pipPos, "slow")
            self.wait([cfut], timeout=None)

            pa.setStatus("Waiting click on pipette")
            pipClickPos = self.getClickPosition()
            pipetteErr = pipClickPos - (targetClickPos + np.array([0, 0, 10e-6]))

            resultText.addRecord(
                {"pipette": pipetteErr, "stage": stageErr,}
            )



import numpy as np
from acq4.util.Mutex import Mutex


class JobQueue(object):
    """Stores a list of jobs and assigns them by request.

    Pipettes are selected for each job based on the quadrant that the target appears in.
    """

    def __init__(self, pipettes, module):
        self.pipettes = pipettes  # {pipette_name: quadrant}
        self.module = module
        self.protocol = None
        self.all_jobs = []
        self.queued_jobs = []
        self.center = module.plateCenter()
        self.enabled = False
        self.positions = np.empty((0, 3))
        self.lock = Mutex(recursive=True)

    def setProtocol(self, prot):
        self.protocol = prot
        self.setJobs(self.all_jobs)

    def setEnabled(self, en):
        """If enabled, then requestJob() will attempt to return the next available job.
        If disabled, then requestJob() will return None.
        """
        self.enabled = en

    def setJobs(self, jobs):
        with self.lock:
            # queue up all jobs that have not run this protocol yet
            self.all_jobs = jobs
            self.queued_jobs = [j for j in jobs if self.protocol.name not in j.assigned_protocols]

    def requestJob(self, pipette):
        # Simple implementation: return the job nearest to the current pipette position.
        # Only jobs in the same quadrant as the pipette are considered.

        with self.lock:
            if not self.enabled:
                return None

            if len(self.queued_jobs) == 0:
                return None

            # current pipette position
            pos = np.array(pipette.pipetteDevice.globalPosition())

            # all job positions
            positions = np.array([job.position for job in self.queued_jobs])

            # which quadrant does this pipette belong in?
            pip_quad = np.array(self.pipettes[pipette.name()]).astype(bool)

            # mask cells that are not in the same quadrant
            quad_center = np.array(self.center[:2])
            cell_quads = positions[:, :2] > quad_center
            quad_mask = (cell_quads == pip_quad[None, :]).all(axis=1)

            # find closest cell to this pipette, excluding other quads
            diff = positions - np.array(pos).reshape(1, 3)
            dist = (diff ** 2).sum(axis=1) ** 0.5
            dist[~quad_mask] = np.inf
            closest = np.argmin(dist)
            if dist[closest] == np.inf:
                return None

            job = self.queued_jobs.pop(closest)
            job.setProtocol(self.protocol)
            job.assignPipette(pipette)
            return job



import numpy as np
from acq4.devices.Sensapex import Sensapex


class AutopatchCustomStage(Sensapex):
    """Implement extra motion planning to protect objective.
    """

    def _move(self, abs, rel, speed, linear, protected=True):
        if not protected:
            return Sensapex._move(self, abs, rel, speed, linear)

        wells = np.array(self.config["wellPositions"]) * 1e9
        radius = self.config["wellRadius"] * 1e9
        wellZ = self.config["wellMaxZ"] * 1e9
        safeZ = self.config["safeMaxZ"] * 1e9

        pos1 = np.array(self.getPosition())
        pos2 = np.array(self._toAbsolutePosition(abs, rel))

        # lateral distances from center of each well
        dist1 = ((wells - pos1[np.newaxis, :2]) ** 2).sum(axis=1) ** 0.5
        dist2 = ((wells - pos2[np.newaxis, :2]) ** 2).sum(axis=1) ** 0.5

        # closest well
        well1 = np.argmin(dist1)
        well2 = np.argmin(dist2)

        # are we starting / ending in the same well?
        startInWell = np.any(dist1 < radius)
        endInWell = np.any(dist2 < radius)
        changeWell = well1 != well2
        # print("startInWell: %s  endInWell: %s   changeWell: %s   wells: %s %s" % (startInWell, endInWell, changeWell, well1, well2))

        # don't move too high
        maxZ = wellZ if endInWell else safeZ
        pos2[2] = min(pos2[2], maxZ)

        if startInWell and endInWell and not changeWell:
            # no danger here, just move
            # print("   no well change; moving directly")
            return Sensapex._move(self, abs=pos2, rel=None, speed=speed, linear=linear)

        # decide on a safe path
        path = []
        # print("stage move path:")

        # First move down to safe Z if needed
        lastZ = pos1[2]
        if pos1[2] > safeZ:
            wp1 = pos1.copy()
            wp1[2] = min(safeZ, pos2[2])
            lastZ = wp1[2]
            path.append({"abs": wp1, "speed": "fast", "protected": False})
            # print("   - move focus down")

        # Next move just XY
        wp2 = pos2.copy()
        wp2[2] = lastZ
        path.append({"abs": wp2, "speed": speed, "protected": False})
        # print("   - move xy")

        # Finally correct Z if needed
        if wp2[2] != pos2[2]:
            path.append({"abs": pos2, "speed": "fast", "protected": False})
            # print("   - move focus up")

        return self.movePath(path)



from __future__ import print_function, division

import os

import pyqtgraph as pg
from acq4.Manager import getManager
from acq4.modules.Module import Module
from acq4.util import Qt
from acq4.util.prioritylock import PriorityLock
from acq4.util.target import Target

from .job_queue import JobQueue
from .patch_attempt import PatchAttempt
from .patch_thread import PatchThread
from .protocols import allPatchProtocols

MainForm = Qt.importTemplate(".main_window")


class AutopatchModule(Module):
    moduleDisplayName = "Autopatch"
    moduleCategory = "Acquisition"

    def __init__(self, manager, name, config):
        # lock used to serialize access to shared stage/camera hardware
        self.stageCameraLock = PriorityLock()
        self._stageLockRequest = None

        self.patchAttempts = []
        self._cammod = None
        self._camdev = None
        self._nextPointID = 0
        self._plateCenter = config.get("plateCenter", (0, 0, 0))

        Module.__init__(self, manager, name, config)

        self.win = Qt.QWidget()
        self.win.resize(1600, 900)
        self.win.closeEvent = self.closeEvent
        self.ui = MainForm()
        self.ui.setupUi(self.win)

        for protocol in allPatchProtocols():
            self.ui.protocolCombo.addItem(protocol)

        for i, w in enumerate([40, 130, 100, 400]):
            self.ui.pointTree.header().resizeSection(i, w)

        # hard-code pipette name => status text box assignments
        # we'd have to rewrite this if we move beyond 4 pipettes or want to rename them..
        self.pipStatusText = {f"PatchPipette{i:d}": getattr(self.ui, f"pip{i:d}Status") for i in range(1, 5)}

        self.win.show()

        self.ui.addPointsBtn.toggled.connect(self.addPointsToggled)
        self.ui.removePointsBtn.clicked.connect(self.removePointsClicked)
        self.ui.startBtn.toggled.connect(self.startBtnToggled)
        self.ui.abortBtn.clicked.connect(self.abortClicked)
        self.ui.resetBtn.clicked.connect(self.resetClicked)
        self.ui.pointTree.itemSelectionChanged.connect(self.treeSelectionChanged)
        self.ui.protocolCombo.currentIndexChanged.connect(self.protocolComboChanged)
        self.ui.lockStageBtn.toggled.connect(self.lockStageBtnToggled)

        camMod = self.getCameraModule()

        pc = self.plateCenter()
        self.plateCenterLines = [
            pg.InfiniteLine(pos=pc[:2], angle=0, movable=False),
            pg.InfiniteLine(pos=pc[:2], angle=90, movable=False),
        ]
        for line in self.plateCenterLines:
            camMod.window().addItem(line)
        radius = 5e-3
        self.wellCircles = [
            Qt.QGraphicsEllipseItem(x - radius, y - radius, radius * 2, radius * 2) for x, y in config["wellPositions"]
        ]
        for wc in self.wellCircles:
            wc.setPen(pg.mkPen("y"))
            camMod.window().addItem(wc)

        cam = self.getCameraDevice()
        cam.sigGlobalTransformChanged.connect(self.cameraTransformChanged)

        # allow to disable safe move in config
        self.safeMoveEnabled = config.get("safeMove", True)

        self.jobQueue = JobQueue(config["patchDevices"], self)

        self.threads = []
        for pipName in config["patchDevices"]:
            pip = manager.getDevice(pipName)
            pip.setActive(True)

            # Write state config parameters to pipette state manager.
            # This does not play nicely with others; perhaps we should have our own state manager.
            stateConfig = pip.stateManager().stateConfig
            for k, v in config.get("patchStates", {}).items():
                stateConfig.setdefault(k, {})
                stateConfig[k].update(v)

            thread = PatchThread(pip, self)
            self.threads.append(thread)
            thread.start()

        self.loadConfig()
        self.protocolComboChanged()

    def window(self):
        return self.win

    def addPointsToggled(self):
        cammod = self.getCameraModule()
        if self.ui.addPointsBtn.isChecked():
            self.ui.startBtn.setChecked(False)
            cammod.window().getView().scene().sigMouseClicked.connect(self.cameraModuleClicked)
        else:
            Qt.disconnect(cammod.window().getView().scene().sigMouseClicked, self.cameraModuleClicked)

    def getCameraModule(self):
        if self._cammod is None:
            manager = getManager()
            mods = manager.listInterfaces("cameraModule")
            if len(mods) == 0:
                raise Exception("Open the Camera module first")
            self._cammod = manager.getModule(mods[0])
        return self._cammod

    def getCameraDevice(self):
        if self._camdev is None:
            manager = getManager()
            camName = self.config.get("imagingDevice", None)
            if camName is None:
                cams = manager.listInterfaces("camera")
                if len(cams) == 1:
                    camName = cams[0]
                else:
                    raise Exception(
                        f"Single camera device required (found {len(cams):d}) or 'imagingDevice' key in configuration."
                    )
            self._camdev = manager.getDevice(camName)
        return self._camdev

    def protocolComboChanged(self):
        prot = str(self.ui.protocolCombo.currentText())
        self.jobQueue.setProtocol(allPatchProtocols()[prot])

    def cameraModuleClicked(self, ev):
        if ev.button() != Qt.Qt.LeftButton:
            return

        camera = self.getCameraDevice()
        cameraPos = camera.mapToGlobal([0, 0, 0])

        globalPos = self._cammod.window().getView().mapSceneToView(ev.scenePos())
        globalPos = [globalPos.x(), globalPos.y(), cameraPos[2]]

        self.addPatchAttempt(globalPos)

    def addPatchAttempt(self, position):
        pid = self._nextPointID
        self._nextPointID += 1

        item = Qt.QTreeWidgetItem([str(pid), "", "", ""])
        self.ui.pointTree.addTopLevelItem(item)

        target = Target(movable=False)
        self._cammod.window().addItem(target)
        target.setPos(pg.Point(position[:2]))
        target.setDepth(position[2])
        target.setFocusDepth(position[2])
        target.circles = []
        for r in (3e-6, 5e-6):
            c = pg.QtGui.QGraphicsEllipseItem(0, 0, 1, 1)
            c.scale(r * 2, r * 2)
            c.setPos(-r, -r)
            c.setPen(pg.mkPen("b"))
            c.setParentItem(target)
            target.circles.append(c)

        pa = PatchAttempt(pid, position, item, target)
        item.patchAttempt = pa
        self.patchAttempts.append(pa)

        self.jobQueue.setJobs(self.patchAttempts)

        pa.statusChanged.connect(self.jobStatusChanged)

        return pa

    def selectedProtocol(self):
        return allPatchProtocols()[str(self.ui.protocolCombo.currentText())]

    def removePointsClicked(self):
        sel = self.ui.pointTree.selectedItems()
        for item in sel:
            self.removePatchAttempt(item.patchAttempt)

    def removePatchAttempt(self, pa):
        self.patchAttempts.remove(pa)

        index = self.ui.pointTree.indexOfTopLevelItem(pa.treeItem)
        self.ui.pointTree.takeTopLevelItem(index)

        pa.targetItem.scene().removeItem(pa.targetItem)

        self.jobQueue.setJobs(self.patchAttempts)

    def cameraTransformChanged(self):
        cam = self.getCameraDevice()
        fdepth = cam.mapToGlobal([0, 0, 0])[2]

        for pa in self.patchAttempts:
            pa.targetItem.setFocusDepth(fdepth)

    def startBtnToggled(self):
        if self.ui.startBtn.isChecked():
            self.ui.startBtn.setText("Stop")
            self.ui.addPointsBtn.setChecked(False)
            self.jobQueue.setJobs(self.patchAttempts)
            self.jobQueue.setEnabled(True)
        else:
            self.ui.startBtn.setText("Start")
            self.jobQueue.setEnabled(False)

    def abortClicked(self):
        """Stop all running jobs.
        """
        self.ui.startBtn.setChecked(False)
        for thread in self.threads:
            thread.stop()
            thread.wait()
            thread.start()

    def resetClicked(self):
        """Reset the state of all points so they can be run again.
        This is mostly meant for development to allow quick iteration.
        """
        for pa in self.patchAttempts:
            pa.reset()
        self.jobQueue.setJobs(self.patchAttempts)

    def closeEvent(self, ev):
        self.quit()
        return Qt.QWidget.closeEvent(self.win, ev)

    def quit(self):
        self.ui.startBtn.setChecked(False)
        self.ui.addPointsBtn.setChecked(False)
        for thread in self.threads:
            thread.stop()
        for pa in self.patchAttempts[:]:
            self.removePatchAttempt(pa)
        for item in self.plateCenterLines + self.wellCircles:
            scene = item.scene()
            if scene is not None:
                scene.removeItem(item)
        self.saveConfig()
        return Module.quit(self)

    def jobStatusChanged(self, job, status):
        item = job.treeItem
        pip = job.pipette.name() if job.pipette is not None else ""
        item.setText(1, "" if job.protocol is None else job.protocol.name)
        item.setText(2, pip)
        item.setText(3, status)
        pipnum = pip[12:] if pip.lower().startswith("patchpipette") else pip
        statusTxt = self.pipStatusText.get(pip)
        if statusTxt is not None:
            statusTxt.setText(f"{pipnum}: {status}")

    def deviceStatusChanged(self, device, status):
        # todo: implement per-pipette UI
        print(f"Device status: {device}, {status}")

    def treeSelectionChanged(self):
        sel = self.ui.pointTree.selectedItems()
        if len(sel) == 1:
            # TODO: something more user-friendly; this is just for development
            log = sel[0].patchAttempt.formatLog()
            self.ui.resultText.setPlainText(log)

    def plateCenter(self):
        return self._plateCenter

    def saveConfig(self):
        geom = self.win.geometry()
        config = {
            # 'plateCenter': list(self._plateCenter),
            # 'window': str(self.win.saveState().toPercentEncoding()),
            "geometry": [geom.x(), geom.y(), geom.width(), geom.height()],
        }
        configfile = os.path.join("modules", self.name + ".cfg")
        man = getManager()
        man.writeConfigFile(config, configfile)

    def loadConfig(self):
        configfile = os.path.join("modules", self.name + ".cfg")
        man = getManager()
        config = man.readConfigFile(configfile)
        if "geometry" in config:
            geom = Qt.QRect(*config["geometry"])
            self.win.setGeometry(geom)
        # if 'window' in config:
        #     ws = Qt.QByteArray.fromPercentEncoding(config['window'])
        #     self.win.restoreState(ws)
        # if 'plateCenter' in config:
        #     self.setPlateCenter(config['plateCenter'])

    def lockStageBtnToggled(self, v):
        if self._stageLockRequest is not None:
            self._stageLockRequest.release()
            self._stageLockRequest.sigFinished.disconnect(self.stageLockAcquired)
            self._stageLockRequest = None

        if v is True:
            # acquire stage lock with higher priority than patch threads
            self._stageLockRequest = self.stageCameraLock.acquire(priority=10)
            self._stageLockRequest.sigFinished.connect(self.stageLockAcquired)
            self.ui.lockStageBtn.setText("Locking stage...")
        else:
            self.ui.lockStageBtn.setText("Lock stage")

    def stageLockAcquired(self, req):
        self.ui.lockStageBtn.setText("Stage locked!")



from os import path

from setuptools import setup, find_packages

packages = [x for x in find_packages(".") if x.startswith("acq4_autopatch")]

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    author="Luke Campagnola",
    author_email="luke.campagnola@gmail.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    description=("Automated cell patching extension for ACQ4"),
    install_requires=[
        "acq4",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="LGPL-3",
    name="acq4_autopatch",
    packages=packages,
    url="https://github.com/sensapex/acq4-autopatch",
    version="0.0.4",
)
